
/*
    VPPP virtual PPP connection over TCP.

    Copyright (C) 1998,1999  Maxim Krasnyansky <max_mk@yahoo.com>

    Idea is based on Sam Lantinga's PPPTCP package, but code was 
    completely written from the scratch.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*/

/*
 * Version: 2.1 03/06/1999 Maxim Krasnyansky <max_mk@yahoo.com>
 */ 

#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <signal.h>
#include <syslog.h>

#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include "vppp.h"

struct vppp_host{
   char host[9];
   char pass[9];
   int speed_in;
   int speed_out;
   char pppd_args[PPPD_ARGS_LEN];
} hosts[VPPP_MAX_HOSTS];
int hosts_num=0;

char *proc_name="vpppd";

char *hosts_file=VPPP_HOSTS_FILE;

int port=VPPP_PORT;

extern char *optarg;
extern int optind,opterr,optopt;

int read_hosts(void);
int auth_host(int fd);
inline int find_host(char *host, char *pass);

void sighup(int sig);

int main(int argc, char *argv[], char *env[]){
     int s,rt_fd;
     int h,opt;
     char *ip;
     struct sigaction sa;
     struct sockaddr_in my_addr,cl_addr;

     while( (opt=getopt(argc,argv,"f:P:")) != EOF ){
	switch(opt){
	    case 'P':
		port=atoi(optarg);
		break;
	    case 'f':
		hosts_file=strdup(optarg);
		break;
	    default:
		printf("Usage: vpppd [-P port] [-f hosts_file]\n");
	        exit(1);
	}
     }	

     /* Become daemon */ 
     if( fork() )
	exit(0);

     close(0);close(1);close(2);
     setsid();

     openlog("vpppd",LOG_PID,LOG_DAEMON);
    
     switch( (hosts_num=read_hosts()) ){
	case 0:
	   syslog(LOG_ERR,"No hosts specified in %s",hosts_file);
	   exit(1);
	case -1:
	   syslog(LOG_ERR,"Error reading %s",hosts_file);
	   exit(1);
     }	
	
     memset(&sa,0,sizeof(sa));     
     sa.sa_handler=sighup;
     sigaction(SIGHUP,&sa,NULL);

     sa.sa_handler=SIG_IGN;
     sigaction(SIGINT,&sa,NULL);
     sigaction(SIGQUIT,&sa,NULL);
     sigaction(SIGCHLD,&sa,NULL);
     sigaction(SIGPIPE,&sa,NULL);

     memset(&my_addr,0,sizeof(my_addr));
     my_addr.sin_family = AF_INET;
     my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
     my_addr.sin_port = htons(port);	

     if( (s=socket(AF_INET,SOCK_STREAM,0))== -1 ){
	syslog(LOG_ERR,"Can't creat socket");
	exit(1);
     }

     opt=1;
     setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)); 

     if( bind(s,(struct sockaddr *)&my_addr,sizeof(my_addr)) ){
	syslog(LOG_ERR,"Can't bind to the socket");
	exit(1);
     }

     if( listen(s,5) ){
	syslog(LOG_ERR,"Can't listen on the socket");
	exit(1);
     }

     init_title(argc,argv,env);
     set_title("%s: waiting for connections on port %d",proc_name,port);

     while(1){
        opt=sizeof(cl_addr);
	if( (rt_fd=accept(s,&cl_addr,&opt)) == -1 )
	   continue; 

	if( !fork() ){
     	   set_title("%s: authentification",proc_name);
	   ip=inet_ntoa(cl_addr.sin_addr);
	   if( (h=auth_host(rt_fd)) != -1 ){	
              sa.sa_handler=SIG_IGN;
              sigaction(SIGHUP,&sa,NULL);

	      syslog(LOG_INFO,"Session %s[%s] speed %d:%d", hosts[h].host,ip,
					hosts[h].speed_out,hosts[h].speed_in);

	      start_ppp(rt_fd,hosts[h].host,hosts[h].speed_in,hosts[h].pppd_args);
	      syslog(LOG_INFO,"Session %s[%s] closed", hosts[h].host, ip);
	   } else {
	      syslog(LOG_INFO,"Denied connection from %s", ip);
	   }
	
	   close(rt_fd);
	   exit(0);
        }
	close(rt_fd);
     }  
}	

void sighup(int sig)
{
     switch( (hosts_num=read_hosts()) ){
	case 0:
	   syslog(LOG_ERR,"No hosts specified in %s",hosts_file);
	   exit(1);
	case -1:
	   syslog(LOG_ERR,"Error rereading %s",hosts_file);
	   exit(1);
     }	
}


int read_hosts(void)
{
     FILE *f;
     char buf[255],*str,*ptr;
     int h=0,line=0,valid=0;

     if((f=fopen(hosts_file,"r"))==NULL)
        return -1;
  
     while( fgets(buf,sizeof(buf),f) ) {
	valid=0;
	line++;

	str=buf;
	while(*str == ' ') str++;

     	if(*str== '#' || *str == '\n'){
	   valid=1; 
	   continue;
	}

	strtok(str,"\n\r");

     	if( (str=strtok(str," \t")) == NULL)
	   break;

	/* host:pass */
	
	if( (ptr=strchr(str,':')) == NULL)
	   break;
	*ptr++='\0';

	if( strlen(str) > sizeof(hosts[0].host)-1 )
	   break;
        strcpy(hosts[h].host,str);

	if( strlen(ptr) > sizeof(hosts[0].pass)-1 )
	   break;
	strcpy(hosts[h].pass,ptr);

	/* flags and/or speed */

 	if( (str=strtok(NULL," \t")) == NULL)
	   break;

       	if( (ptr=strchr(str,'C')) )
	  hosts[h].speed_out = hosts[h].speed_in = -1;
	else if( (ptr=strchr(str,':')) ){
	  *ptr++='\0';
          hosts[h].speed_out=atoi(ptr);
          hosts[h].speed_in=atoi(str);
	} else 
	  hosts[h].speed_out = hosts[h].speed_in = atoi(str);

	/* pppd_args */

 	if( (str=strtok(NULL,"\t")) == NULL)
	   break;
  	while(*str == ' ') str++;
     	if( strlen(str) > sizeof(hosts[0].pppd_args)-1 )
	   break;
        strcpy(hosts[h].pppd_args,str);      	

        valid=1;
	if(++h > VPPP_MAX_HOSTS){
	  syslog(LOG_ERR,"Maximum number of hosts excided");
	  fclose(f);
	  return -1;
	} 
     }
     if(!valid){
       syslog(LOG_ERR,"Incorrect options on the line %d",line);
       fclose(f);
       return -1; 
     }
     return h;	
 }

#define ST_HOST  1
#define ST_PASS  2

int auth_host(int fd){
	int stage,h=-1;
	FILE *fl;
	char buf[255],*str1,*str2;
	char host[9],pass[9];	

	fl=fdopen(dup(fd),"w");
	setvbuf(fl,NULL,_IONBF,0);

	fprintf(fl,"VPPP server %s\n",VERS);

	stage = ST_HOST;

	while( read_t(fd,buf,sizeof(buf),VPPP_TIMEOUT) > 0){
	   buf[sizeof(buf)-1]='\0';
	   strtok(buf,"\r\n");

	   if( (str1=strtok(buf," :")) == NULL )
		break;
	   if( (str2=strtok(NULL," :")) == NULL )
		break;

	   switch(stage){
	     case ST_HOST:
	        if( !strncmp(str1,"HOST",4) ){
		  strncpy(host,str2,sizeof(host));
		  host[sizeof(host)-1]='\0';
		  stage = ST_PASS;
		  fprintf(fl,"OK\n");
		  continue;
	        }
		break;
	     case ST_PASS:
	        if( !strncmp(str1,"PASS",4) ){
		  strncpy(pass,str2,sizeof(pass));
		  pass[sizeof(pass)-1]='\0';
		  if( (h=find_host(host,pass)) != -1 ){
		     fprintf(fl,"OK %d\n",hosts[h].speed_out);
		     break;
		  }
	        }
		break;
 	   }
	   break;
	}

	if(h == -1)
	  fprintf(fl,"ERR\n");	

	fclose(fl);

	return h;
}

 
inline int find_host(char *host, char *pass){
	int i;

	for(i=0; i < hosts_num; i++)
	   if( !strcmp(host,hosts[i].host) && !strcmp(pass,hosts[i].pass ) )
		return i; 
	
	return -1;
}
	

